import type { Storage } from "./Storage";
import type { Moment } from "moment";
import { v4 as uuidv4 } from "uuid";
import { EventEmitter } from "./event";

export interface ReadingLogEntry {
	id: string;
	book: string;
	pagesRead: number;
	pagesReadTotal: number;
	pagesRemaining: number;
	createdAt: Moment;
}

interface ReadingLogEventMap {
	load: { entries: ReadingLogEntry[] };
	created: { entry: ReadingLogEntry };
	updated: { entry: ReadingLogEntry };
	removed: { entry: ReadingLogEntry };
}

export class ReadingLog extends EventEmitter<ReadingLogEventMap> {
	private entries: ReadingLogEntry[] = [];

	public constructor(private readonly storage: Storage) {
		super();

		this.load().catch((error) => {
			console.error("Failed to load reading log entries:", error);
		});
	}

	async load(filename = "reading-log.json") {
		const entries = await this.storage.readJSON<ReadingLogEntry[]>(
			filename
		);
		if (entries) {
			this.entries = entries.map((entry) => ({
				...entry,
				// @ts-expect-error Moment is provided by Obsidian
				createdAt: moment(entry.createdAt),
			}));
			this.emit("load", { entries: this.entries });
		}
	}

	private sortEntries() {
		this.entries = this.entries.sort((a, b) =>
			a.createdAt.diff(b.createdAt)
		);
	}

	async save(filename = "reading-log.json") {
		this.sortEntries();
		await this.storage.writeJSON(
			filename,
			this.entries.map((entry) => ({
				...entry,
				// @ts-expect-error Moment is provided by Obsidian
				createdAt: moment(entry.createdAt).toISOString(true),
			}))
		);
	}

	public getEntries(): ReadingLogEntry[] {
		return this.entries;
	}

	public getLastEntryForBook(book: string): ReadingLogEntry | null {
		const entriesForBook = this.entries.filter(
			(entry) => entry.book === book
		);

		return entriesForBook.length > 0
			? entriesForBook[entriesForBook.length - 1]
			: null;
	}

	public async createEntry(
		book: string,
		pageEnded: number,
		pageCount: number
	): Promise<void> {
		const lastEntry = this.getLastEntryForBook(book);

		if (lastEntry && lastEntry.pagesReadTotal >= pageEnded) {
			throw new Error(
				"Given page ended is less than the previous entry's pages read total."
			);
		}

		const newEntry: ReadingLogEntry = {
			id: uuidv4(),
			book,
			pagesRead: lastEntry
				? pageEnded - lastEntry.pagesReadTotal
				: pageEnded,
			pagesReadTotal: pageEnded,
			pagesRemaining: pageCount - pageEnded,
			// @ts-expect-error Moment is provided by Obsidian
			createdAt: moment(),
		};

		if (
			lastEntry &&
			lastEntry.createdAt.isSame(newEntry.createdAt, "day")
		) {
			newEntry.id = lastEntry.id;
			newEntry.pagesRead += lastEntry.pagesRead;
			await this.updateEntry(newEntry);
		} else {
			await this.addRawEntry(newEntry);
		}
	}

	public async addRawEntry(entry: ReadingLogEntry) {
		this.entries.push(entry);
		await this.save();
		this.emit("created", { entry });
	}

	public async updateEntry(entry: ReadingLogEntry): Promise<void> {
		this.entries[this.entries.findIndex((other) => other.id === entry.id)] =
			entry;
		await this.save();
		this.emit("updated", { entry });
	}

	public async removeEntry(entry: ReadingLogEntry): Promise<void> {
		const index = this.entries.findIndex((other) => other.id === entry.id);

		if (index !== -1) {
			const removed = this.entries.splice(index, 1);
			await this.save();
			this.emit("removed", { entry: removed[0] });
		}
	}

	public async removeEntriesForBook(book: string): Promise<void> {
		this.entries = this.entries.filter((entry) => entry.book !== book);
		await this.save();
	}

	public async clearEntries(): Promise<void> {
		this.entries = [];
		await this.save();
	}
}
